@@ -5,8 +5,7 @@
 # This script is largely based on the Fluent Linter used in mozilla-central
 # https://firefox-source-docs.mozilla.org/code-quality/lint/linters/fluent-lint.html
 
-from fluent.syntax import parse, visitor
-from fluent.syntax import serializer
+from fluent.syntax import ast, parse, serializer, visitor
 from html.parser import HTMLParser
 import argparse
 import bisect
@@ -94,6 +93,12 @@ class Linter(visitor.Visitor):
             "can_have_group_comment": True,
             # If currently looking at a term
             "is_term": False,
+            # Comment bound to the current message
+            "comment": "",
+            # The current group comment
+            "group_comment": "",
+            # Variables in the current message
+            "variables": [],
         }
 
         # Set this to true to debug print the root node's json. This is useful for
@@ -116,7 +121,6 @@ class Linter(visitor.Visitor):
         return False
 
     def check_typography(self, node):
-
         # Serialize message without comments
         parts = []
         parts.append(f"{node.id.name} =")
@@ -247,9 +251,31 @@ class Linter(visitor.Visitor):
 
         # Check typography
         self.check_typography(node)
-
         super().generic_visit(node)
 
+        if "VC" in self.config and not self.config["VC"]["disabled"]:
+            # Check if variables are referenced in comments
+            if self.state["variables"]:
+                comments = self.state["comment"] + self.state["group_comment"]
+                missing_references = [
+                    v for v in self.state["variables"] if f"${v}" not in comments
+                ]
+                if missing_references:
+                    self.add_error(
+                        node,
+                        node.id.name,
+                        "VC01",
+                        "Messages including variables should have a comment "
+                        "explaining what will replace the variable. "
+                        "This can be either a message or group comment. "
+                        "\n            Missing references: "
+                        + ", ".join([f"${m}" for m in missing_references]),
+                    )
+
+        # Reset comment and variable references after reading the message
+        self.state["comment"] = ""
+        self.state["variables"] = []
+
     def visit_MessageReference(self, node):
         # Log errors if message references are not supported
         if "SY02" in self.config and self.config["SY02"]["disabled"]:
@@ -258,13 +284,14 @@ class Linter(visitor.Visitor):
         # We don't recurse into message references, the identifiers are either
         # checked elsewhere or are attributes and come from DOM.
 
-        pass
-
     def visit_TermReference(self, node):
         # Log errors if term references are not supported
         if "SY03" in self.config and self.config["SY03"]["disabled"]:
             self.add_error(node, None, "SY03", "Terms are not supported.")
-        pass
+
+        # Reset comment and variable references after reading the message
+        self.state["comment"] = ""
+        self.state["variables"] = []
 
     def visit_TextElement(self, node):
         html_stripper = MLStripper()
@@ -378,6 +405,35 @@ class Linter(visitor.Visitor):
             )
             return
 
+    def visit_Placeable(self, node):
+        if (
+            "PS01" in self.config
+            and not self.config["PS01"]["disabled"]
+            and type(node.expression) not in [ast.SelectExpression]
+        ):
+            if type(node.expression) == ast.VariableReference:
+                placeable_name = f"${node.expression.id.name}"
+            elif type(node.expression) == ast.TermReference:
+                placeable_name = f"-{node.expression.id.name}"
+            else:
+                placeable_name = node.expression.id.name
+            if (node.span.start + 2) != node.expression.span.start:
+                self.add_error(
+                    node,
+                    self.last_message_id,
+                    "PS01",
+                    f"Placeables should be preceded by exactly one space (`{{ {placeable_name} }}`).",
+                )
+            if (node.span.end - 2) != node.expression.span.end:
+                self.add_error(
+                    node,
+                    self.last_message_id,
+                    "PS01",
+                    f"Placeables should be followed by exactly one space (`{{ {placeable_name} }}`).",
+                )
+
+        super().generic_visit(node)
+
     def visit_SelectExpression(self, node):
         # Log errors if variants are not supported
         if node.variants and "SY04" in self.config and self.config["SY04"]["disabled"]:
@@ -389,9 +445,26 @@ class Linter(visitor.Visitor):
             for variant in node.variants:
                 super().generic_visit(variant.value)
 
+        # Store the variable used for the SelectExpression, excluding functions
+        # like PLATFORM()
+        if (
+            type(node.selector) == ast.VariableReference
+            and node.selector.id.name not in self.state["variables"]
+        ):
+            self.state["variables"].append(node.selector.id.name)
+
+    def visit_Comment(self, node):
+        # This node is a comment with: "#"
+
+        # Store the comment
+        self.state["comment"] = node.content
+
     def visit_GroupComment(self, node):
         # This node is a comment with: "##"
 
+        # Store the group comment
+        self.state["group_comment"] = node.content
+
         # Skip if checks for group comments are disabled
         if "GC" in self.config and self.config["GC"]["disabled"]:
             return
@@ -402,7 +475,7 @@ class Linter(visitor.Visitor):
                 None,
                 "GC04",
                 "Group comments (##) must be followed by at least one message. Make sure "
-                "that a single group comment with multiple pararaphs is not separated by "
+                "that a single group comment with multiple paragraphs is not separated by "
                 "whitespace, as it will be interpreted as two different comments.",
             )
             return
@@ -451,12 +524,13 @@ class Linter(visitor.Visitor):
         # We don't recurse into variable references, the identifiers there are
         # allowed to be free form.
 
+        if node.id.name not in self.state["variables"]:
+            self.state["variables"].append(node.id.name)
+
         # Log errors if variable references are not supported
         if "SY06" in self.config and self.config["SY06"]["disabled"]:
             self.add_error(node, None, "SY06", "Variable references are not supported.")
 
-        pass
-
     def add_error(self, node, message_id, rule, msg):
         (col, line) = self.span_to_line_and_col(node.span)
 
@@ -525,7 +599,6 @@ def get_newlines_count_before(span, contents):
 
 
 def lint(file_paths, config_path):
-
     # Get list of FTL files
     files = {}
     total_files = 0
